// Copyright (c) 2024 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "ShaderLoader.hpp"
#include "assert/assert.hpp"
#include "asset/AssetManager.hpp"
#include "game/Game.hpp"

using namespace Dawn;

const std::string ShaderLoader::ASSET_TYPE = "shader";

ShaderLoader::ShaderLoader(
  const std::shared_ptr<AssetManager> assetManager,
  const std::string name
) :
  AssetLoader(assetManager, name),
  state(ShaderLoaderState::INITIAL)
{
}

void ShaderLoader::updateAsync() {
}

void ShaderLoader::updateSync() {
  if(state != ShaderLoaderState::INITIAL) return;

  this->state = ShaderLoaderState::LOADING;
  assertFalse(loaded, "ShaderLoader already loaded.");

  auto sm = this->getAssetManager()->getGame()->shaderManager;

  // Load the shader string
  Slang::ComPtr<IBlob> diagnostics;
  module = sm->session->loadModule(
    this->name.c_str(),
    diagnostics.writeRef()
  );

  // Get list of entry points and create components
  int32_t definedEntryPointCount = module->getDefinedEntryPointCount();
  IComponentType** components = new IComponentType*[definedEntryPointCount + 1];
  int32_t componentCount = 0;

  components[componentCount++] = module;

  for(auto i = 0; i < definedEntryPointCount; i++) {
    Slang::ComPtr<IEntryPoint> ep;
    auto result = module->getDefinedEntryPoint(i, ep.writeRef());

    if(result != SLANG_OK) {
      assertUnreachable("Failed to get entry point.");
      return;
    }

    auto name = ep->getFunctionReflection()->getName();
    std::cout << "Found entry point: " << name << std::endl;
    entryPoints.push_back(std::string(name));
    components[componentCount++] = ep;
  }

  // Create the composite component type
  sm->session->createCompositeComponentType(
    components,
    componentCount,
    program.writeRef(),
    diagnostics.writeRef()
  );
  if(diagnostics) {
    assertUnreachable("%s\n", (const char*) diagnostics->getBufferPointer());
    return;
  }

  // Link the program.
  auto result = program->link(linkedProgram.writeRef(), diagnostics.writeRef());
  if(diagnostics) {
    assertUnreachable("%s\n", (const char*) diagnostics->getBufferPointer());
    return;
  }

  delete [] components;
  this->state = ShaderLoaderState::LOADED;
  this->loaded = true;
}

std::string ShaderLoader::getAssetType() const {
  return ShaderLoader::ASSET_TYPE;
}

std::string ShaderLoader::getEntryPointCode(const std::string &entryPoint) {
  assertTrue(loaded, "ShaderLoader not loaded.");
  assertNotNull(linkedProgram, "ShaderLoader linkedProgram is null.");

  // Get the entry point index
  int32_t entryIndex = -1;
  for(auto i = 0; i < entryPoints.size(); i++) {
    if(entryPoints[i] != entryPoint) continue;
    entryIndex = i;
    break;
  }
  assertTrue(entryIndex != -1, "EntryPoint not found.");

  // Find the entry point code
  Slang::ComPtr<IBlob> blob;
  Slang::ComPtr<IBlob> diagnostics;
  auto result = linkedProgram->getEntryPointCode(
    entryIndex,
    0,
    blob.writeRef(),
    diagnostics.writeRef()
  );

  if(diagnostics) {
    assertUnreachable("%s\n", (const char*) diagnostics->getBufferPointer());
    return "";
  }

  return std::string((const char*)blob->getBufferPointer());
}

ShaderLoader::~ShaderLoader() {
  if(linkedProgram) {
    linkedProgram->release();
    linkedProgram = nullptr;
  }
}