// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "display/shader/Shader.hpp"
#include "ShaderManagerSlangFileSystem.hpp"
using namespace slang;

namespace Dawn {
  class Game;
  class ShaderLoader;

  class ShaderManager : public std::enable_shared_from_this<ShaderManager> {
    private:
      std::vector<std::weak_ptr<IShaderBase>> shaders;

      std::weak_ptr<Game> game;
      ShaderManagerSlangFileSystem fileSystem;

      Slang::ComPtr<IGlobalSession> globalSession;
      TargetDesc targetDescription;
      SessionDesc sessionDescription;
      Slang::ComPtr<ISession> session;

    public:
      /**
       * Creates a new shader manager.
       */
      ShaderManager();

      /**
       * Initializes the shader manager.
       * 
       * @param game The game instance that the shader manager is being used in.
       */
      void init(const std::shared_ptr<Game> &game);

      /**
       * Retreives the game instance.
       * 
       * @return Game instance.
       */
      std::shared_ptr<Game> getGame();

      /**
       * Retreives an instance of the shader from the shader manager. If the
       * shader does not exist it will be created.
       * 
       * @tparam T Type of shader to retreive.
       * @return Shader instance.
       */
      template<class T>
      std::shared_ptr<T> getShader() {
        auto itShaders = shaders.begin();
        while(itShaders != shaders.end()) {
          auto shader = itShaders->lock();
          if(!shader) {
            itShaders = shaders.erase(itShaders);
            continue;
          }

          std::shared_ptr<T> casted = std::dynamic_pointer_cast<T>(shader);
          if(casted) return casted;
          itShaders++;
        }

        auto newShader = std::make_shared<T>();
        shaders.push_back(newShader);
        newShader->init();
        return newShader;
      }

      /**
       * Disposes of all shaders.
       */
      ~ShaderManager();

      friend class ShaderLoader;
  };
}