// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "Scene.hpp"
#include "assert/assert.hpp"

using namespace Dawn;

Scene::Scene(const std::shared_ptr<Game> &game) : game(game) {
  assertNotNull(game, "Game cannot be NULL");
}

void Scene::init() {
  assertFlagOff(state, SceneState::INITIALIZED, "Scene already initialized");
  Flag::turnOn(state, SceneState::INITIALIZED);

  for(auto &item : sceneItems) {
    item->init();
  }
}

void Scene::deinit() {
  assertFlagOn(state, SceneState::INITIALIZED, "Scene not initialized");
  assertFlagOff(state, SceneState::DEINITIALIZED,"Scene already deinitialized");

  auto items = this->sceneItems;
  for(auto &item : items) {
    item->deinit();
  }

  sceneItems.clear();
  sceneItemsToRemove.clear();

  Flag::turnOn(state, SceneState::DEINITIALIZED);
}

void Scene::update() {
  // Initialize new scene items
  assertFlagOn(state, SceneState::INITIALIZED, "Scene not initialized");
  
  // Remove stale scene items.
  auto itRemove = sceneItemsToRemove.begin();
  while(itRemove != sceneItemsToRemove.end()) {
    auto item = *itRemove;
    item->deinit();
    auto it = std::find(sceneItems.begin(), sceneItems.end(), item);
    if(it != sceneItems.end()) {
      sceneItems.erase(it);
    }
    itRemove++;
  }
  sceneItemsToRemove.clear();

  // Tick scene items.
  onNextFrame.emit();
  onNextFrame.clear();

  float_t delta = getGame()->timeManager.delta;
  if(paused) {
    onPausedUpdate.emit(delta);
  } else {
    onUnpausedUpdate.emit(delta);
    
    onTimeout.tick(delta);
    onInterval.tick(delta);
  }
}

std::shared_ptr<Game> Scene::getGame() {
  auto game = this->game.lock();
  assertNotNull(game, "Game is NULL");
  return game;
}

std::shared_ptr<SceneItem> Scene::createSceneItem() {
  assertFlagOff(state, SceneState::DEINITIALIZED, "Scene deinitialized");
  auto item = std::make_shared<SceneItem>(shared_from_this());
  sceneItems.push_back(item);
  return item;
}

void Scene::removeItem(const std::shared_ptr<SceneItem> item) {
  assertFlagOn(state, SceneState::INITIALIZED, "Scene not initialized");
  assertFlagOff(state, SceneState::DEINITIALIZED, "Scene deinitialized");

  auto index = std::find(sceneItems.begin(), sceneItems.end(), item);
  if(index == sceneItems.end()) return;
  
  sceneItemsToRemove.push_back(item);
}

Scene::~Scene() {
  if(Flag::isOn(state, SceneState::INITIALIZED)) {
    assertFlagOn(state, SceneState::DEINITIALIZED, "Scene not deinitialized!");
  }
}