// Copyright (c) 2024 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "scene/SceneItem.hpp"
#include "dawnphysics.hpp"

namespace Dawn {
  enum class ColliderType {
    DYNAMIC,
    STATIC,
    KINEMATIC
  };

  class Collider : public SceneComponent {
    private:
      JPH::Body *body = nullptr;
      ColliderType colliderType = ColliderType::DYNAMIC;

      /**
       * Returns the JoltPhysics motion type for the collider type.
       * 
       * @param colliderType The collider type.
       * @return The JoltPhysics motion type.
       */
      static JPH::EMotionType getMotionType(const ColliderType colliderType);

    protected:
      JPH::BodyID bodyId;
      JPH::ObjectLayer layer = 1;

      /**
       * Returns the shape settings for the collider.
       * 
       * @return The shape settings for the collider.
       */
      virtual std::shared_ptr<JPH::ShapeSettings> getShapeSettings() = 0;

      /**
       * Returns the JoltPhysics body interface system.
       * 
       * @return The JoltPhysics body interface system.
       */
      JPH::BodyInterface & getBodyInterface();

      /**
       * Callable by subclasses to notify that the shape has changed.
       */
      void notifyShapeChanged();

    public:
      void onInit() override;
      void onDispose() override;

      /**
       * Returns whether the collider is ready.
       * 
       * @return Whether the collider is ready.
       */
      bool_t isColliderReady();

      /**
       * Returns the collider type.
       * 
       * @return The collider type.
       */
      ColliderType getColliderType();

      /**
       * Returns the JoltPhysics body ID of the collider.
       * 
       * @return The body ID of the collider.
       */
      JPH::BodyID getBodyId();

      /**
       * Returns the velocity of the collider.
       * 
       * @return The velocity of the collider.
       */
      glm::vec3 getVelocity();

      /**
       * Sets the collider type.
       * 
       * @param colliderType The collider type.
       */
      void setColliderType(ColliderType colliderType);

      /**
       * Sets the velocity of the collider.
       * 
       * @param velocity The velocity.
       */
      void setVelocity(const glm::vec3 velocity);

      /**
       * Adds a force to the collider.
       * 
       * @param force Force to add.
       * @param inPoint Application point of the force.
       */
      void addForce(
        const glm::vec3 force,
        const glm::vec3 inPoint = glm::vec3(0, 0, 0)
      );
  };
}